/*
 * MPC Feedback Module 1 for FATOM Windows (analysis)
 *
 * Copyright (C) 2000-2004 by Grigori G. Fursin
 *
 * homepage: http://homepages.inf.ed.ac.uk/gfursin
 *
 * ICSA, School of Informatics,
 * University of Edinburgh, UK
 */

#include <stdio.h>
#include <malloc.h>
#include <assert.h>

static int proc_max=0;
static int inst_max=0;

static long* RegAccess;
static long* RegAccessC;
static long* RegAccessRW;
static unsigned char* AddrValue;

static int first_run=0;

void mpc_init(int proc_max1, int inst_max1)
{
  long x;
  long x1;
  long y;

  proc_max=proc_max1;
  inst_max=inst_max1;

  x1=proc_max*inst_max;
  x=(4*x1);
  printf("FATOM (eos_mpc_fb1): Allocating %u bytes of memory [RegAccess]...\n", x);

  RegAccess=(long *) calloc(x1, 4);

  if (RegAccess == NULL) 
  {
    printf("\nFatal Error (FATOM, eos_mpc_fb1): Can't allocate memory!\n");
    exit(1);
  }

  for (y=0; y<x1; y++)
  {
    RegAccess[y]=0;
  }

  x1=proc_max*inst_max;
  x=(4*x1);
  printf("FATOM (eos_mpc_fb1): Allocating %u bytes of memory [RegAccessC] ...\n", x);

  RegAccessC=(long *) calloc(x1, 4);

  if (RegAccess == NULL) 
  {
    printf("\nFatal Error (FATOM, eos_mpc_fb1): Can't allocate memory!\n");
    exit(1);
  }

  for (y=0; y<x1; y++)
  {
    RegAccessC[y]=0;
  }

  x1=proc_max*inst_max;
  x=(4*x1);
  printf("FATOM (eos_mpc_fb1): Allocating %u bytes of memory [RegAccessRW]...\n", x);

  RegAccessRW=(int *) calloc(x1, 4);

  if (RegAccessRW == NULL) 
  {
    printf("\nFatal Error (FATOM, eos_mpc_fb1): Can't allocate memory!\n");
    exit(1);
  }

  for (y=0; y<x1; y++)
  {
    RegAccessRW[y]=0;
  }

  x1=proc_max*inst_max*10;
  x=(1*x1);
  printf("FATOM (eos_mpc_fb1): Allocating %u bytes of memory [AddrValue]...\n", x);

  AddrValue=(char *) calloc(x1, 1);

  if (AddrValue == NULL) 
  {
    printf("\nFatal Error (FATOM, eos_mpc_fb1): Can't allocate memory!\n");
    exit(1);
  }

  for (y=0; y<x1; y++)
  {
    AddrValue[y]=0;
  }

  printf("\nFATOM (eos_mpc_fb1): Allocated successfully!\n\n");
}

int mpc_finish(void)
{
  FILE *f;
  int i=0;
  int j=0;
  int k=0;
  char fname[1024];

  unsigned int val=0;
  int q=0;

  printf("\nFATOM (eos_mpc_fb1): Saving collected data ...\n");

  for (i=0; i<proc_max; i++)
  {
    sprintf(fname, "_eos_mpc_atom_tmp5.%u", i);

    if ((f=fopen(fname,"wt"))==NULL)
    {
      printf("\nFatal Error (FATOM, eos_mpc_fb1): Can't write file!\n");
      exit(1);
    }

    for (k=0; k<inst_max; k++)
    {
      if (RegAccessC[i*inst_max+k]!=0)
      {
        fprintf(f, "inst={%5u}, change={%3u}, addr={%16ld} (0x%08lx), rw={%1u}, value={...} (0x", 
                   k, 
                   RegAccessC[i*inst_max+k],
                   RegAccess[i*inst_max+k],
                   RegAccess[i*inst_max+k],
                   RegAccessRW[i*inst_max+k]);

        for (j=0; j<10; j++)
        {
          val=AddrValue[(i*inst_max+k)*10+9-j]*256+AddrValue[(i*inst_max+k)*10+9-j-1];
          j++;
          fprintf(f, "%04hx", val);
        }

        fprintf(f, ")\n");
      }
      else
      {
        fprintf(f, "inst={%5u}, change={%3u}, addr={%16ld} (0x%08lx), rw={%1u}, value={...} (0x0)\n", 
                   k, 
                   RegAccessC[i*inst_max+k],
                   0,
                   0,
                   0);
      }
    }
   
    fclose(f);
  }

  printf("FATOM (eos_mpc_fb1): Data saved successfully!\n");

  return 0;
}

void mpc_reg_access(long addr, long proc, long inst, long rw, unsigned char* addr1, unsigned char* addr2)
{
  int q=0;
  unsigned char* a=0;
 
  if (first_run==0)
  {
    mpc_init1();

    first_run=1;
  }

  if (RegAccessC[proc*inst_max+inst]==2) 
  {
    /* Removing call by nops */
    for (a=addr1; a<addr2; a++) *a=0x90;
    return;
  }
  else if (RegAccessC[proc*inst_max+inst]==1)
  {
    if (RegAccess[proc*inst_max+inst]!=addr)
    {
      RegAccessC[proc*inst_max+inst]=2;
    }
  }
  else
  {
    RegAccessC[proc*inst_max+inst]=1;
    RegAccess[proc*inst_max+inst]=addr;
    RegAccessRW[proc*inst_max+inst]=rw;

    if (rw==0)
    {
      for (q=0; q<10; q++) AddrValue[(proc*inst_max+inst)*10+q]=* (char*) (addr+q);
    }
  }
}
