/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// $Id: Expressions.cc,v 1.14 2000/01/07 12:45:15 fgg Exp $

#include "Octave/Base/ParseContext.h"
#include "Octave/Base/Unparse.h"
#include "Octave/Fortran/Fortran.h"

// Here are some useful macros to simplify the code for checking
// operand types in expressions.

#define VOID_TYPE(_e_) \
 (((_e_) && (_e_)->type) ? (_e_)->type->isBuiltInType(T_VOID,0) : 0)

#define INTEGRAL(_e_) \
 (((_e_) && (_e_)->type) ? (_e_)->type->isIntegralType() : 0)

#define ARITHMETIC(_e_) \
 (((_e_) && (_e_)->type) ? (_e_)->type->isArithType() : 0)

#define AGGREGATE(_e_) \
 (((_e_) && (_e_)->type) ? (_e_)->type->isAggregateType() : 0)

#define POINTER(_e_) \
 (((_e_) && (_e_)->type) ? (_e_)->type->isPointerType() || \
 (_e_)->type->isArrayType() : 0)

#define ARITH_OR_PTR(_e_) \
 (((_e_) && (_e_)->type) ? (_e_)->type->isArithType() || (_e_)->type->isPointerType() \
 || (_e_)->type->isArrayType() : 0)

#define FUNC_RETURN_TYPE(_t_) \
  ((_t_) ? (_t_)->node.fields.base_type : 0)

#define POINTER_BASE_TYPE(_t_) \
  ((_t_) ? (_t_)->node.fields.base_type : 0)

// Static data...

static oprec cops[] = {
  { " == ",      10 },
  { " < ",       11 },
  { " > ",       11 },
  { " != ",      10 },
  { " <= ",      11 },
  { " >= ",      11 },
  { " + ",       13 },
  { " - ",       13 },
  { " || ",	5 },
  { " * ",       14 },
  { " / ",       14 },
  { " % ",       14 },
  { " && ",	6 },
  { "EXP_OP",	0 },
  { "ARRAY_MULT",0 },
  { "CONCAT_OP",0 },
  { " ^ ",	8 },
  { " == ",      10 },
  { " != ",      10 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { " - ",       13 },
  { " ! ",       15 },   /* NOT_OP     */
  { " = ",	3 },
  { " * ",       15 },   /* DEREF_OP   */
  { "->",      16 },   /* POINTST_OP */
  { "FUNCTION_OP",16 },
  { "--",      15 },   /* increment if postfix */
  { "++",      15 },   /* increment if postfix */
  { " & ",	9 },   /* BITAND_OP  */
  { " | ",	7 },   /* BITOR_OP   */
  { "STAR_RANGE", 0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "PROC_CALL",0 },
  { "FUNC_CALL",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "UNUSED_OP",0 },
  { "CONSTRUCTOR_REF",0},
  { ".",       16 },      /* ACCESS_REF */
  { "CONS",	0 },
  { "ACCESS",	0 },
  { "IOACCESS",	0 },
  { "CONTROL_LIST",0},
  { "SEQ",	0 },
  { "SPEC_PAIR",0 },
  { "COMM_LIST",0 },
  { "STMT_STR",	0 },
  { "EQUI_LIST",0 },
  { "IMPL_TYPE",0 },
  { "STMTFN_DECL",0},
  { "~",       15 },     /* BIT_COMPLEMENT_OP */
  { " ? ",	4 },
  { " : ",	2 },
  { "FUNCTION_REF",20},
  { " << ",      12 },
  { " >> ",      12 },
  { " + ",       13 },
  { "sizeof",	0 },
  { " / ",       14 },
  { " - ",       13 },
  { " <= ",      11 },
  { " >= ",      11 },
  { " != ",      10 },
  { "CLASSINIT_OP",0},
  { "CAST_OP",	0 },
  { "&",       15 },     /* ADDRESS_OP */
  { "POINTSAT_OP",0},
  { "COPY_NODE",0 },
  { "INIT_LIST",	0 },
  { "VECTOR_CONST",	0 },
  { "BIT_NUMBER",	0 },
  { "ARITH_ASSGN_OP",	0 },
  { "ARRAY_OP",	0 },
  { "new",	0 },
  { "delete",	0 },
  { "NAMELIST_LIST",	0 },
  { "this",	0 },
  { "::",	0 },
  { " += ",	3 },
  { " -= ",	3 },
  { " &= ",	3 },
  { " |= ",	3 },
  { " *= ",	3 },
  { " /= ",	3 },
  { " %= ",	3 },
  { " ^= ",	3 },
  { " <<= ",	3 },
  { " >>= ",	3 }
};

// Static function forward-declarations...

static char* cop_name (expression_tag) ;
static int cop_prec (expression_tag) ;

// Member functions...

/*--------------------------------------------------------------------+
 |                                                                    |
 |               Expression::set_fields (Expression&)                 |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Expression::set_fields (Expression& e)
{
  pc = e.context() ;
  expr_variant (e.expr_variant()) ;
  type = e.type ;
  node.dval = e.node.dval ;
  node.fields.symbol = e.node.fields.symbol ;
  if (e.node.fields.e1)
    node.fields.e1 = pc->make_expr (*(e.node.fields.e1)) ;
  else
    node.fields.e1 = 0 ;
  if (e.node.fields.e2)
    node.fields.e2 = pc->make_expr (*(e.node.fields.e2)) ;
  else
    node.fields.e2 = 0 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                     Expression::set_expr_type                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Expression::set_expr_type ()
{
  int r_type ;
  int l_type ;
  PTR_TYPE temp, l_operand, r_operand ;
  
  switch (variant()) {
    
  case (ARRAY_MULT):	/* to be changed */
  case (PROC_CALL):
  case (DEF_CHOICE):
  case (VAR_LIST):
  case (EXPR_LIST):
  case (RANGE_LIST):
  case (RANGE_OP):
  case (STAR_RANGE):
    type = pc->global_default;
    break;
    
  case (INT_VAL):
    type = pc->global_int;
    break;

  case (FLOAT_VAL):
    type = pc->global_float;
    break;

  case (DOUBLE_VAL):
    type = pc->global_double;
    break;

  case (BOOL_VAL):
    type = pc->global_bool;
    break;

  case (CHAR_VAL):
    type = pc->global_char;
    break;

  case (STMT_STR):
  case (STRING_VAL):
    type = pc->global_string;
    break;

  case (COMPLEX_VAL):
    type = pc->global_complex;
    break;

  case (CONST_REF):	/* check that the symbol referred to is a
			 * constant and then return type */
    if (node.const_ref.symbol->variant() != CONST_NAME) {
      pc->error ("constant reference does not have a constant id");
      type = pc->global_default;
    } 
    else
      type = node.const_ref.symbol->type;
    break;

  case (ENUM_REF):
  case (VAR_REF):
    {
      // Set type of expression to be the type of the referenced symbol.
      Symbol *sym = node.var_ref.symbol ;
      type = sym->type;

      // If this reference is to an external, ensure it has at least
      // a non-null type (give it 'default' type) and return.

      if (sym->attr & EXTERNAL_BIT) {
	if (!type)
	  type = pc->global_default ;
	return ;
      }

      // All other references must be to symbols with a defined type.
      // TBD: this needs to be tidied up to cope with the problem
      //      in Fortran, when a function name is used as a variable
      //
      // if (!type || type == pc->global_default)
      //   pc->error ("identifier not declared");
      
      if (!type)
	type = global_default ; // This will not always be correct
      break;
    }
    
  case (ARRAY_REF):	/* check to see if symbol is array */
    if (type->type_variant() == T_ARRAY)
      type = type->node.ar_decl.base_type ;
    break;
    
  case (RECORD_REF):
    type = node.record_ref.rec_field ?
      node.record_ref.rec_field->type :
	node.record_ref.symbol->type;
    break;

  case (MOD_OP):
  case (LOWER_OP):
  case (UPPER_OP):
    type = pc->global_int;
    break;

  case (AND_OP):
  case (OR_OP):
  case (EQ_OP):
  case (LT_OP):
  case (GT_OP):
  case (NOTEQL_OP):
  case (LTEQL_OP):
  case (EQV_OP):
  case (NEQV_OP):
  case (GTEQL_OP):
/*
    // fgg remarked those lines on 2000/11/30 - later needs TBD?
    if ((node.binary_op.l_operand == 0) ||
	(node.binary_op.l_operand->type == 0) ||
	(node.binary_op.r_operand == 0) ||
	(node.binary_op.r_operand->type == 0))
      pc->error ("inconsistent operands to boolean operation.");
    // fgg changed next line on 2000/11/30 -> was just if
*/
    if ((node.binary_op.l_operand!=0) && 
        (node.binary_op.l_operand->type!=0) && 
        (node.binary_op.l_operand->type->variant()==T_ARRAY))
    {
      type = new Datatype (*(node.binary_op.l_operand->type)) ;
      type->node.ar_decl.base_type = pc->global_bool ;
    }
    else
    {
      type = pc->global_bool;
    }
    break;

  case (DIV_OP):
  case (ADD_OP):
  case (SUBT_OP):
  case (MULT_OP):
    /* When correct type attributions are working re-enable this check --- TBD ---
    if ((node.binary_op.l_operand == 0) ||
	(node.binary_op.l_operand->type == 0) ||
	(node.binary_op.r_operand == 0) ||
	(node.binary_op.r_operand->type == 0))
      pc->error ("inconsistent operands to arithmetic operation");
    */
    l_operand = node.binary_op.l_operand->type;
    r_operand = node.binary_op.r_operand->type;
    if (! l_operand || ! r_operand)
      type = pc->global_default;
    else {
      if (l_operand->variant() == T_ARRAY)
	l_type = l_operand->node.ar_decl.base_type->variant();
      else
	l_type = l_operand->variant();
      if (r_operand->variant() == T_ARRAY)
	r_type = r_operand->node.ar_decl.base_type->variant();
      else
	r_type = r_operand->variant();

      if (l_type == T_COMPLEX || r_type == T_COMPLEX)
	{
	  type = pc->global_complex ;
	  // The following check is strict F77
	  // if (l_type == T_DOUBLE || r_type == T_DOUBLE)
	  //   pc->error ("cannot mix COMPLEX and DOUBLE PRECISION operands") ;
	}
      else if (l_type == T_DOUBLE || r_type == T_DOUBLE)
	type = pc->global_double;
      else if (l_type == T_FLOAT || r_type == T_FLOAT)
	type = pc->global_float;
      else if (l_type == T_INT && r_type == T_INT)
	type = pc->global_int;
      else 
	type = pc->global_default;
    }
    break;
  case (NOT_OP):
  case (UNARY_ADD_OP):
  case (MINUS_OP):
    type = node.unary_op.operand->type;
    break;
  case (EXP_OP):
    type = node.binary_op.l_operand->type;
    break;
  case (CONCAT_OP):
    type = node.binary_op.l_operand->type;
    break;
  case (DDOT):
    type = node.binary_op.r_operand->type;
    break;

  case (FUNC_CALL):
    type = node.fields.symbol->type ;
    break ;

  //fgg added on 2000/12/11 to deal with (,) correctly
  case (PAREN_OP):
    type = node.binary_op.l_operand->type;
    break;

  default:
    pc->error ("internal: expression unclassifiable");
    break;
    
  }
}



/*--------------------------------------------------------------------+
 |                                                                    |
 |                      Expression::match_an_arg                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

// Was: match_an_arg (actual_arg, formal_arg)
int Expression::match_an_arg (PTR_SYMB formal_arg)
{
// Was: PTR_EXPR actual_arg;
// Was: PTR_SYMB formal_arg;
  
  PTR_TYPE actual_arg_type, formal_arg_type ;
  
  actual_arg_type = this->type ;
  formal_arg_type = formal_arg->type ;
  
  if (actual_arg_type->variant() != formal_arg_type->variant()) 
    return 0 ;
  
  if ((actual_arg_type->variant() == T_DERIVED_TYPE) &&
      (actual_arg_type->name != formal_arg_type->name))
    return 0 ;
  
  if (actual_arg_type->variant() == T_ARRAY)
  {
    if (actual_arg_type->node.ar_decl.base_type->variant() !=
	formal_arg_type->node.ar_decl.base_type->variant())
      return 0 ;
    
    if ((actual_arg_type->node.ar_decl.base_type->variant() == T_DERIVED_TYPE)
	&& (actual_arg_type->node.ar_decl.base_type->name !=
	    formal_arg_type->node.ar_decl.base_type->name))
      return 0 ;
    
    if (actual_arg_type->node.ar_decl.num_dimensions !=
	formal_arg_type->node.ar_decl.num_dimensions)
      return 0 ;
  }
  
  formal_arg->decl = 1 ;
  return 1 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                 Expression::is_array_section_ref                   |
 |                                                                    |
 +--------------------------------------------------------------------*/

int Expression::is_array_section_ref ()
{
  int num_triplets = 0 ;
  PTR_EXPR p ;
  PTR_TYPE t ;
  
  p = this ;
  while (p) 
  {
    // check subscript range
    if (p->node.fields.e1->variant() == DDOT) num_triplets++ ;

    if ((t = p->node.fields.e1->type))
    {
      if (t->variant() == KEYWORD_ARG) 
	return 0 ;

      // check vector subscript
      if ((t->variant() == T_ARRAY) 
	  && (t->node.ar_decl.base_type->variant() == T_INT))
	num_triplets++ ;
    }
    p = p->node.fields.e2 ;
  }
  
  return (num_triplets) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   Expression::is_substring_ref                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

int Expression::is_substring_ref ()
{
  PTR_EXPR l1 = this ? this->node.fields.e1 : 0 ;

  if (l1 && (l1->variant() == KEYWORD_ARG) )
    return 0 ;

  if (this
      && (this->node.fields.e2 == 0) 
      && l1 
      && (l1->variant() == DDOT)
      && (l1->node.fields.e1->variant() == DDOT))
    return 1 ;
  else 
    return 0 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       Expression::base_type                        |
 |                                                                    |
 +--------------------------------------------------------------------*/

type_tag Expression::base_type ()
{
  if (!type)
    return T_DEFAULT ;
  
  type_tag t = type->type_variant() ;

  if (t == T_ARRAY)
    return type->node.ar_decl.base_type->type_variant() ;
  else
    return type->type_variant() ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                         Expression::typing                         |
 |                                                                    |
 +--------------------------------------------------------------------*/

type_tag Expression::typing ()
{
  if (!type)
    return T_DEFAULT ;
  else
    return type->type_variant() ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |            Expression::coerceToType (PTR_CNTXT, PTR_TYPE)          |
 |                                                                    |
 +--------------------------------------------------------------------*/

char* Expression::coerceToType (PTR_TYPE t, char *action)
{
  if (t->isPointerType())
    {
      // attempt pointer conversion from this->type to t
      int ev = 0 ;
      int isCst ;
      if (POINTER(this))
	{
	  if (POINTER_BASE_TYPE(type)->type_variant() == T_VOID)
	    return 0 ;
	  if (*t != *type)
	    return type_error (action, type, t) ;
	}
      else if (type && type->type_variant() == T_FUNCTION)
	{
	  // this is a reference to a function.
	  // We can convert implicitly from function to pointer-to-function
	  // but should check that their return types are equivalent.
	  
	  PTR_TYPE rt = FUNC_RETURN_TYPE(type) ;
	  PTR_TYPE lt = FUNC_RETURN_TYPE(POINTER_BASE_TYPE(t)) ;
	  if (!rt || !lt || *(rt) != *(lt))
	    {
	      // function is not assignment compatible with pointer
	      return type_error (action, type, t) ;
	    }
	}
      else if (!(isCst = isIntConst(&ev)) || (isCst && ev))
	return type_error (action, type, t) ;
    }
  else if (t->isArithType() && ARITHMETIC(this))
    {
      // attempt arithmetic conversion from type to t
      // TBD: 
      // For the moment assume this is OK.
    }
  else if (t->isAggregateType() && AGGREGATE(this))
    {
      if (*type != *t)
	return type_error (action, type, t) ;
    }
  else if (t->isIntegralType() && POINTER(this))
      pc->warn ("conversion from pointer to integer without a cast") ;
  else
    return type_error (action, type, t) ;
 
 return 0 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       Expression::isIntConst                       |
 |                                                                    |
 +--------------------------------------------------------------------*/

bool Expression::isIntConst (int *i)
{
  if (!this) return false ;

  expression_tag et = expr_variant() ;
  int x, y ;
  bool ix, iy ;

  switch (et)
    {
    case INT_VAL:
      *i = node.ival ;
      return true ;
//      break ;

    case BOOL_VAL:
      *i = (int)(node.bval) ;
      return true ;
//      break ;

    case CHAR_VAL:
      *i = (int)(node.cval) ;
      return true ;
//      break ;

    case CONST_REF:
      {
	PTR_SYMB csym = node.fields.symbol ;
	PTR_EXPR cexp = csym->node.const_value ;
	if (cexp)
	  return cexp->isIntConst (i) ;
	else
	  return false ;
//	break ;
      }

    case VAR_REF:
      {
	PTR_SYMB vs = node.fields.symbol ;
	if (!vs)
	  return false ;

	PTR_TYPE vt = vs->type ;
	if (!vt)
	  return false ;

	symbol_tag ss = vs->symb_variant() ;

	if (ss == CONST_NAME && vs->node.const_value)
	  return vs->node.const_value->isIntConst (i) ;
	else if (ss == VARIABLE_NAME && (vt->type_flags & GOT_CONST))
	  return vs->node.const_value->isIntConst (i) ;
	else
	  return false ;
//	break ;
      }
    case EQ_OP:
    case EQV_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x == y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case LT_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x < y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case GT_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x > y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case LE_OP:
    case LTEQL_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x <= y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case GE_OP:
    case GTEQL_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x >= y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case NE_OP:
    case NOTEQL_OP:
    case NEQV_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x != y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case ADD_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x + y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case SUBT_OP:
    case SUB_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x - y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case MULT_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x * y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case DIV_OP:
    case INTEGER_DIV_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y) && y)
	{
	  *i = x / y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case MOD_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y) && y)
	{
	  *i = x % y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case OR_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x || y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case AND_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x && y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case XOR_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x ^ y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case BITAND_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x & y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case BITOR_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x | y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case LSHIFT_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x << y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case RSHIFT_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y))
	{
	  *i = x >> y ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case MINUS_OP:
      if (node.fields.e1->isIntConst(&x))
	{
	  *i = -x ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case NOT_OP:
      if (node.fields.e1->isIntConst(&x))
	{
	  *i = ~x ;
	  return true ;
	}
      else
	return false ;
//      break ;

    case UNARY_ADD_OP:
    case SIZE_OP:
    case CAST_OP:   // perhaps the target type should be integer?
      if (node.fields.e1->isIntConst(&x))
	{
	  *i = x ;
	  return true ;
	}
      else
	return false ;
//      break ;
      
    case EXP_OP:
      if (node.fields.e1->isIntConst(&x) && node.fields.e2->isIntConst(&y) && y >= 0)
	{
	  if (y)
	    *i = x ;
	  else
	    *i = 1 ;
	  if (y > 1 && x != 1)
	    while (y)
	      {
		*i = *i * x ;
                y--;
	      }
	  return true ;
	}
      else
	return false ;
//      break ;

//    default:
//      return false ;
  }

  return false ; //fgg added on 2000.12.16
}

bool Expression::hasNodeASymbol ()
{
  if (!this)
    return false ;
  expression_tag etag = expr_variant() ;
  switch (etag)
    {
    case VAR_REF:
    case ARRAY_REF:
    case PROC_CALL:
    case TYPE_REF:
    case FUNC_CALL:
    case FUNCTION_REF:
      return true ;
    default: return false ;
    }
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |           Expression::FtnUnparseExpression (LineBuffer&)            |
 |                                                                     |
 +---------------------------------------------------------------------*/

void Expression::FtnUnparseExpression (LineBuffer &lb)
{
  FortranUnparseExpression (lb, this) ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                 static void CcUnparseExpression                     |
 |                                                                     |
 +---------------------------------------------------------------------*/

void Expression::CcUnparseExpression (SourceListing& src, int par_prec)
{
  if (!this)
    return ;
  
  expression_tag etag = expr_variant() ;
  int prec = cop_prec (etag) ;

  if (par_prec > prec)
    src << "(" ;

  switch (etag)
    {
      // Atomic expressions
      //
    case INT_VAL:
      {
	// Note: an integer value may be `int' `long' or `unsigned long'
	char sb[64];
	if (type == pc->global_int)
	  sprintf (sb, "%d", node.ival);
	else if (type == pc->global_long_int)
	  sprintf (sb, "%ld", node.lval);
	else if (type == pc->global_unsigned_int)
	  sprintf (sb, "%u", node.uval);
	else if (type == pc->global_unsigned_long_int)
	  sprintf (sb, "%lu", node.uval);
	src << sb ;
	break;
      }
    
    case FLOAT_VAL:
    case DOUBLE_VAL:
      src << (*node.string_val)() ;
      break;
    case STRING_VAL:
      src << '"' ;
      src << (*node.string_val)() ;
      src << '"' ;
      break;

    case BOOL_VAL:
      if (node.bval)
	src << "true" ;
      else 
	src << "false" ;
      break;

    case CHAR_VAL:
      switch (node.cval)
	{
	case '\0': src << "'\0'" ; break ;
	case '\a': src << "'\a'" ; break ;
	case '\b': src << "'\b'" ; break ;
	case '\t': src << "'\t'" ; break ;
	case '\n': src << "'\n'" ; break ;
	case '\v': src << "'\v'" ; break ;
	case '\f': src << "'\f'" ; break ;
	case '\r': src << "'\r'" ; break ;
	case '\\': src << "'\\'" ; break ;
	default:
	  {
	    char sbuf [64] ;
	    if (node.cval < 32)
	      {
		sprintf (sbuf, "%03o'", node.cval) ;
		src << '\'' ;
		src << '\\' ;
	      }
	    else
	      sprintf (sbuf, "'%c'", node.cval) ;
	    src << sbuf ;
	    break ;
	  }
	}
      break;
      
    case TYPE_REF:
      {
	UnpCntl upc = node.type_expr.dtag == TYPE_DEFINITION ? UNP_VERBOSE : UNP_BRIEF ;
	node.type_expr.dest_type->CcUnparseType (src, upc) ;
	break ;
      }
      
    case INIT_DECLARATOR:
      node.declarator_list.declarator->CcUnparseDeclarator (src) ;
      if (node.declarator_list.init_expr)
	{
	  src << " = " ;
	  node.declarator_list.init_expr->CcUnparseExpression (src, 0) ;
	}
      break ;

    case CONST_REF:
      src << node.fields.symbol->ident () ;
      break;
      
    case VAR_REF:
    case FUNCTION_REF:
      src << node.fields.symbol->ident () ;
      break ;
      
    case ARRAY_REF:
      node.fields.e1->CcUnparseExpression (src, prec) ;
      src << "[" ;
      node.fields.e2->CcUnparseExpression (src, prec) ;
      src << "]" ;
      break ;
      
    case FUNCTION_OP:
      node.fields.e1->CcUnparseExpression (src, prec) ;
      src << "(" ;
      node.fields.e2->CcUnparseExpression (src, prec) ;
      src << ")" ;
      break;
      
    case EXPR_LIST:
      node.fields.e1->CcUnparseExpression (src, prec) ;
      if (node.fields.e2)
	{
	  src << ", " ;
	  node.fields.e2->CcUnparseExpression (src, prec) ;
	}
      break;
      
      // Left-associative binary operators
      //
    case EQV_OP:
    case NEQV_OP:
    case EQ_OP:
    case LT_OP:
    case GT_OP:
    case NOTEQL_OP:
    case LTEQL_OP:
    case GTEQL_OP:
    case ADD_OP:
    case SUBT_OP:
    case OR_OP:
    case MULT_OP:
    case DIV_OP:
    case MOD_OP:
    case AND_OP:
    case EXP_OP:
    case POINTST_OP:
    case ACCESS_REF:
    case BITAND_OP:
    case BITOR_OP:
      node.binary_op.l_operand->CcUnparseExpression (src, prec) ;
      src << cop_name (etag) ;
      node.binary_op.r_operand->CcUnparseExpression (src, prec-1) ;
      break;
      
      // Prefix OR postfix unary operators
      //
    case PLUSPLUS_OP:
    case MINUSMINUS_OP:
      {
	PTR_EXPR p = node.binary_op.l_operand ;
	if (p)
	  { /* postfix ++ or --, so increment the precedence */
	    p->CcUnparseExpression (src, prec+1) ;
	    src << cop_name (etag) ;
	  }
	else
	  { /* prefix ++ or -- */
	    src << cop_name (etag) ;
	    node.binary_op.r_operand->CcUnparseExpression (src, prec) ;
	  }
	break;
      }
    
      // Unary operators (prefix only)
      //
    case ADDRESS_OP:
    case UNARY_ADD_OP:
    case MINUS_OP:
    case NOT_OP:
    case BIT_COMPLEMENT_OP:
    case DEREF_OP:
      {
	src << cop_name (etag) ;
	node.unary_op.operand->CcUnparseExpression (src, prec) ;
	break;
      }
    
    // Right-associative binary operators
    //
    case ASSGN_OP:
    case AND_ASSGN_OP:
    case IOR_ASSGN_OP:
    case XOR_ASSGN_OP:
    case PLUS_ASSGN_OP:
    case MINUS_ASSGN_OP:
    case LSHIFT_ASSGN_OP:
    case RSHIFT_ASSGN_OP:
    case DIV_ASSGN_OP:
    case MOD_ASSGN_OP:
    case MULT_ASSGN_OP:
      node.fields.e1->CcUnparseExpression (src, prec+1);
      src << ' ' ;
      src << cop_name (etag) ;
      src << ' ' ;
      node.fields.e2->CcUnparseExpression (src, prec);
      break ;
      
    case EXPR_IF:
      {
	PTR_EXPR condition = node.fields.e1 ;
	PTR_EXPR body = node.fields.e2 ;
	condition->CcUnparseExpression (src, prec+1);
	src << " ? " ;
	body->node.fields.e1->CcUnparseExpression (src, prec);
	src << " : " ;
	body->node.fields.e2->CcUnparseExpression (src, prec);
	break ;
      }

    case STRUCTURE_CONSTRUCTOR:
      {
	src << "{ " ;
	node.fields.e1->CcUnparseExpression (src, 0) ;
	src << " }" ;
	break ;
      }

    case SIZE_OP:
      {
	src << "sizeof(" ;
	node.fields.e1->CcUnparseExpression (src, 0) ;
	src << ")" ;
	break ;
      }

    case CAST_OP:
      {
	src << "(" ;
	node.fields.e1->CcUnparseExpression (src, 0) ;
	src << ")" ;
	node.fields.e2->CcUnparseExpression (src, 0) ;
	break ;
      }

    default :
      char sbuf[64] ;
      sprintf (sbuf, " $%d ", expr_variant()) ;
      src << sbuf ;
      break;
    }
  if (par_prec > prec)
    src << ")" ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                     static int cop_prec                             |
 |                                                                     |
 +---------------------------------------------------------------------*/

static int cop_prec (expression_tag etag)
{
  if (etag >= EQ_OP && etag <= RSHIFT_ASSGN_OP)
    return cops[etag - EQ_OP].precedence ;
  else
    return HYPER_PREC ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                     static int cop_name                             |
 |                                                                     |
 +---------------------------------------------------------------------*/

static char* cop_name (expression_tag etag)
{
  if (etag >= EQ_OP && etag <= RSHIFT_ASSGN_OP)
    return cops[etag - EQ_OP].opname ;
  else
    return 0 ;
}

