//
// Class CProf for EOS profile files
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

//Import libraries...
import java.io.*;

public class CProf
{
  private String err;

  private String fname;
  private int    fmax;

  private int n1;
  private int n2;

  private boolean[] q0;
  private String[]  q1;
  private double[]  q2;
  private double[]  q3;
  private double[]  q4;

  //****************************************************************************
  //"Initialization"
  //
  //Input:  fname1 = Prof filename
  //        fmax1  = Max number of active lines in the tree file
  //
  //Return: 0  - Ok
  //        >0 - Error
  //

  public int Init(String fname1,
                  int fmax1)
  {
    int i=0;

    if (fmax1<1) return 1;
    fmax=fmax1;

    fname=fname1;

    n1=0;
    n2=0;

    q0=new boolean [fmax];
    q1=new String  [fmax];
    q2=new double  [fmax];
    q3=new double  [fmax];
    q4=new double  [fmax];

    return 0;
  }

  //****************************************************************************
  //"Parse Prof File"
  //
  //Return: 0  - Ok
  //        <0 - Error
  //
  public int Parse()
  {
    err="";

    BufferedReader fin=null;

    String str1="";

    int ind1=0;
    int ind2=0;

    String str1a="";
    String str1b="";
    String str1c="";
    String str1d="";

    n1=0;
    n2=0;

    //Reading prof
    try
    {
      fin=new BufferedReader(new FileReader(fname));
    }
    catch (IOException e2)
    {
      err="Error: Can't find prof file !";
      return -1;
    }

    try
    {
      while ((str1=fin.readLine())!=null)
      {
        boolean incorrect=true;

        ind1=str1.indexOf("'");
        ind2=str1.indexOf("'", ind1+1);

        if ((ind1!=-1) && (ind2!=-1))
        {
          str1a=str1.substring(ind1+1, ind2);
        
          ind1=str1.indexOf("'", ind2+1);
          ind2=str1.indexOf("'", ind1+1);

          if ((ind1!=-1) && (ind2!=-1))
          {
            str1b=str1.substring(ind1+1, ind2);

            ind1=str1.indexOf("'", ind2+1);
            ind2=str1.indexOf("'", ind1+1);

            if ((ind1!=-1) && (ind2!=-1))
            {
              str1c=str1.substring(ind1+1, ind2);

              ind1=str1.indexOf("'", ind2+1);
              ind2=str1.indexOf("'", ind1+1);

              if ((ind1!=-1) && (ind2!=-1))
              {
                str1d=str1.substring(ind1+1, ind2);

                if (n1>fmax)
                {
                  err="Error: Too many entries in the prof file!";
                  return -1;
                }

                if (str1.charAt(0)=='+')
                {
                  q0[n1]=true;
                  n2++;
                }
                else q0[n1]=false;

                q1[n1]=str1a;

                double d2=0;
                double d3=0;
                double d4=0;
                try
                {
                  d2=Double.valueOf(str1b).doubleValue();
                  d3=Double.valueOf(str1c).doubleValue();
                  d4=Double.valueOf(str1d).doubleValue();
                }
                catch (NumberFormatException nfe1)
                {
                  err="Error: Format of the prof file is invalid!";
                  return -1; 
                }

                q2[n1]=d2;
                q3[n1]=d3;
                q4[n1]=d4;

                n1++;
                         
                incorrect=false;
              }
            }
          }
        }

        if (incorrect)
        {
          err="Error: Can't understand format of the prof file!";
          return -1;
        }
      }
    }
    catch (Exception e1)
    {
      try
      {
        fin.close();
      }
      catch (Exception e2)
      {}
      err="Error: Problems while reading prof file!";
      return -1;
    }

    try
    {
      fin.close();
    }
    catch (Exception e3)
    {}

    return 0;
  }

  //****************************************************************************
  public String GetLastError()
  {
    String err1=err;
    err="";
    return err1;
  }

  //****************************************************************************
  public int GetNumTotal()
  {
    return n1;
  }

  //****************************************************************************
  public int GetNumSelected()
  {
    return n2;
  }

  //****************************************************************************
  public String GetProc(int i, boolean selected)
  {
    if (selected)
    {
      int j=0;
      int k=0;
      for (k=0; k<n1; k++)
      {
        if (q0[k]==true)
        {
          if (j==i) break;
          else j++;
        }
      }
      return q1[k];
    }
    else
    {
      return q1[i];
    }
  }

  //****************************************************************************
  public double GetP1(int i, boolean selected)
  {
    if (selected)
    {
      int j=0;
      int k=0;
      for (k=0; k<n1; k++)
      {
        if (q0[k]==true)
        {
          if (j==i) break;
          else j++;
        }
      }
      return q2[k];
    }
    else
    {
      return q2[i];
    }
  }

  //****************************************************************************
  public double GetP2(int i, boolean selected)
  {
    if (selected)
    {
      int j=0;
      int k=0;
      for (k=0; k<n1; k++)
      {
        if (q0[k]==true)
        {
          if (j==i) break;
          else j++;
        }
      }
      return q3[k];
    }
    else
    {
      return q3[i];
    }
  }

  //****************************************************************************
  public double GetP3(int i, boolean selected)
  {
    if (selected)
    {
      int j=0;
      int k=0;
      for (k=0; k<n1; k++)
      {
        if (q0[k]==true)
        {
          if (j==i) break;
          else j++;
        }
      }
      return q4[k];
    }
    else
    {
      return q4[i];
    }
  }
}
