//
// Class CTransformations for EOS transformations
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

//Import libraries...
import java.io.*;

public class CTransformations
{
  private String err;
  private String f_tmp1="_eos_s_tmp1";
  private String f_tmp2="_eos_s_tmp2";
  private String f_add="_tmp";

  private String f_prj1="_eos_s_prj_tmp1";
  private String f_prj2="_eos_s_prj_tmp2";

  private int cur;
  private int max_instr;
  private CTree ctr;
  private CClient cl;
  private CClient cl1;
  private CProject cp;
  private CTime ct;
  private String str_dir;
  private String s_opt;
  private String s_rt;

  private String f_tmp;

  private String[]  file;
  private String[]  proc;
  private int[]     type1;
  private int[]     type2;
  private String[]  name;
  private String[]  prm1;
  private String[]  prm2;
  private boolean[] corr;

  private String[] filex;

  //****************************************************************************
  //Initialization
  //
  public int Init(int max_instr1, CTree ctr1, CClient clx, CProject cp1,
                  CGlobal g,
                  String str_dir1, String str_fp, String f_tmpx,
                  String s_opt1, String str_rt)
  {
    int ret=0;

    cur=0;

    max_instr=max_instr1;
    ctr=ctr1;
    cl=clx;
    cp=cp1;
    str_dir=str_dir1;
    s_opt=s_opt1;
    f_tmp=f_tmpx;
    s_rt=str_rt;

    if (max_instr<1) 
    { 
      err="Error: Incorrect parameter 'Max Instructions'!";
      return -1;
    }
                        
    file  = new String  [max_instr];
    proc  = new String  [max_instr];
    type1 = new int     [max_instr];
    type2 = new int     [max_instr];
    name  = new String  [max_instr];
    prm1  = new String  [max_instr];
    prm2  = new String  [max_instr];
    corr  = new boolean [max_instr];

    filex = new String [max_instr];

    //Preparing objects
    ct=new CTime();

    //Compiling original application
    cl.Add("initiate compilation",   true, true);
    cl.Add("directory="+str_dir,     true, true);
    cl.Add("file_project="+str_fp,   true, true);
    cl.Add("file_diag_out="+f_tmp,   true, true);
    cl.Add("compilation_mode=0",     true, true);
    cl.Add("no_checking=0",          true, true);
    cl.Add("optimizations="+s_opt,   true, true);
    cl.Add("compile",                true, true);
    cl.Add("terminate",             false, true);

    ret=cl.Send();
    if (ret==-1)
    {
      err="Error: Can't find host with Compilation Server!";
      return -1;
    }
    else if (ret==-2)
    {
      err="Error: Can't connect to Compilation Server!";
      return -1;
    }
    else if (ret==-4)
    {
      err="Error: Can't close connection with Compilation Server!";
      return -1;
    }
    else if (ret<0)
    {
      err=cl.GetLastError();
      return -1;
    }

    //Running
    cl.Add("initiate compilation",   true, true);
    cl.Add("directory="+str_dir,     true, true);
    cl.Add("file_project="+str_fp,   true, true);
    cl.Add("file_out_time="+f_tmp1,  true, true);
    cl.Add("runmode=0",              true, true);
    cl.Add("run",                    true, true);
    cl.Add("terminate",             false, true);

    ret=cl.Send();
    if (ret==-1)
    {
      err="Error: Can't find host with Compilation Server!";
      return -1;
    }
    else if (ret==-2)
    {
      err="Error: Can't connect to Compilation Server!";
      return -1;
    }
    else if (ret==-4)
    {
      err="Error: Can't close connection with Compilation Server!";
      return -1;
    }
    else if (ret<0)
    {
      err=cl.GetLastError();
      return -1;
    }

    //Get time
    if (ct.Init(str_dir+f_tmp1)<0)
    {
      err=ct.GetLastError();
      return -1;
    }

    //Preparing Client for Transformations (Main Language)
    int mainLng=cp.GetLng(0);

    //Create CClient for Transformation Server
    if (g.lng.GetConfig(mainLng).GetParamStr(g.lngTSerHost).equals(""))
    {
      err="Error: Main language in project file is not supported by Transformation Server!";
      return -1;
    }

    cl1=new CClient();
    if (cl1.Init(g.lng.GetConfig(mainLng).GetParamStr(g.lngTSerHost),
                 g.lng.GetConfig(mainLng).GetParamInt(g.lngTSerPort),
                 8192)!=0)
    {
      err="Error (internal): Can't initialize object CClient!";
      return -1;
    }

    return 0;
  }

  //****************************************************************************
  //Add transformation into queue
  //
  //type1x  = 0 - array_common
  //          1 - array
  //          2 - loop
  //type2x  = 0 - array padding
  //          1 - array gir
  //          2 - loop tiling
  //          3 - loop distribution
  //          4 - loop unrolling
  //
  //correct = true - correct loops
  //
  public int Add(int type1x, int type2x, int p1, int i1,
                 String prm1x, String prm2x, boolean correct)
  {
    String file1, proc1, name1;

    if (type1x==0)
    {
      file1=ctr.GetFileAC(i1);
      proc1=ctr.GetProcAC(i1);
      name1=ctr.GetNameAC(i1);
    }
    else if (type1x==1)
    {
      file1=ctr.GetFileA(p1, i1);
      proc1=ctr.GetProcA(p1, i1);
      name1=ctr.GetNameA(p1, i1);
    }
    else
    {
      file1=ctr.GetFileL(p1, i1);
      proc1=ctr.GetProcL(p1, i1);
      name1=ctr.GetCountL(p1, i1);
    }

    if (type1x==0)
    {
      //searching for common arrays in other procedures
      for (int i=0; i<ctr.GetNumArraysCommAll(); i++)
      {
        String file1q=ctr.GetFileACAll(i);
        String proc1q=ctr.GetProcACAll(i);
        String name1q=ctr.GetNameACAll(i);

        if (name1q.equals(name1) && !proc1q.equals(proc1))
        {
          if (cur>=max_instr)
          {
            err="Error: Instruction buffer is full!";
            return -1;
          }

          file[cur] =file1q;
          proc[cur] =proc1q;
          type1[cur]=1;
          type2[cur]=type2x;
          name[cur] =name1q;
          prm1[cur] =prm1x;
          prm2[cur] =prm2x;
          corr[cur] =correct;

          cur++;
        }
      }

      type1x=1;
    }

    if (cur>=max_instr)
    {
      err="Error: Instruction buffer is full!";
      return -1;
    }

    file[cur] =file1;
    proc[cur] =proc1;
    type1[cur]=type1x;
    type2[cur]=type2x;
    name[cur] =name1;
    prm1[cur] =prm1x;
    prm2[cur] =prm2x;
    corr[cur] =correct;

    cur++;

    return 0;
  }

  //****************************************************************************
  public String GetLastError()
  {
    String err1=err;
    err="";
    return err1;
  }

  //****************************************************************************
  public int GetQueuePosition()
  {
    return cur;
  }

  //****************************************************************************
  public void RestoreQueuePosition(int cur1)
  {
    cur=cur1;
  }

  //****************************************************************************
  //Input:  transform =  true - do transformations
  //                    false - just parse (to check variation of running time)
  //Return: -1 - Error (in err)
  //         0 - Success
  //         1 - exe was not created after transformation
  //         2 - output files are incorrect after running transformed version
  //         3 - transformation server returned error
  public int Transform(boolean transform, int step)
  {
    int ret=0;

    if (cur<=0)
    {
      err="Error: Transformation Instruction buffer is empty!";
      return -1;
    }

    int i=0;

    //Making list of files
    String fil="";
    int i_fil=0;

    for (i=0; i<cur; i++)
    {
      boolean new1=true;

      for (int k=0; k<i_fil; k++)
      {
        if (file[i].equals(filex[k]))
        {
          new1=false;
          break;
        }
      }
      
      if (new1)
      {
        filex[i_fil]=file[i];
        i_fil++;
      }
    }

    //Save project with original files only
    String f1=str_dir+f_prj1;

    CProject cp_new = new CProject();
    cp_new.Init(cp, f1);

    cp_new.AddFromProject(filex, i_fil, cp, "");

    if (cp_new.Save()<0)
    {
      err=cp_new.GetLastError();
      return -1;
    }

    //Remove original object files
    cl.Add("initiate compilation",   true, true);
    cl.Add("directory="+str_dir,     true, true);
    cl.Add("file_project="+f_prj1,   true, true);
    cl.Add("remove_obj",             true, true);
    cl.Add("terminate",             false, true);

    ret=cl.Send();
    if (ret==-1)
    {
      err="Error: Can't find host with Compilation Server!";
      return -1;
    }
    else if (ret==-2)
    {
      err="Error: Can't connect to Compilation Server!";
      return -1;
    }
    else if (ret==-4)
    {
      err="Error: Can't close connection with Compilation Server!";
      return -1;
    }
    else if (ret<0)
    {
      err=cl.GetLastError();
      return -1;
    }

    //New project with transformed files
    String f2=str_dir+f_prj2;

    CProject cp_new1 = new CProject();
    cp_new1.Init(cp, f2);

    cp_new1.AddFromProject(filex, i_fil, cp, f_add);

    if (cp_new1.Save()<0)
    {
      err=cp_new1.GetLastError();
      return -1;
    }

    //Navigating through files to transform
    for (i=0; i<i_fil; i++)
    {
      if (cp_new.GetLng(i+1)!=cp_new.GetLng(0))
      {
        err="Error: Language of file(s) is not supported for transformations!";
        return -1;
      }

      boolean record=false;
      BufferedWriter fout=null;

      if (!s_rt.equals(""))
      {
         String x1=CMisc.align_str1(String.valueOf(step), 5, '0');
         String f=str_dir+s_rt+"."+x1+"."+
                 cp_new.GetFileName(i+1)+cp_new.GetFileExt(i+1);

        try
        {
          fout=new BufferedWriter(new FileWriter(f));
        }
        catch (IOException e1)
        {
          err="Can't open file to record transformations!";
          return -1;
        }

        record=true;
      }

      //Build list of transformation
      String q1=cp_new.GetFileName(i+1)+cp_new.GetFileExt(i+1);
      String q2=cp_new1.GetFileName(i+1)+cp_new1.GetFileExt(i+1);

      cl1.Add("initiate transformations",  true, true);
      cl1.Add("filein="+str_dir+q1,        true, true);
      cl1.Add("fileout="+str_dir+q2,       true, true);
      cl1.Add("fileouterror="+f_tmp2,      true, true);

      if (record)
      {
        try
        {
          fout.write("initiate transformations");
          fout.newLine();
          fout.write("filein="+str_dir+q1);
          fout.newLine();
          fout.write("fileout="+str_dir+q2);
          fout.newLine();
          fout.write("fileouterror="+f_tmp2);
          fout.newLine();
        }
        catch (Exception e2)
        {
          err=("Error: Problems while recording transformations into file!");
          return -1;
        }
      }

      if (transform)
      {
        //Choose all transformations for this file
        for (int j=0; j<cur; j++)
        {
          if (file[j].equals(q1))
          {
            cl1.Add("procedure="+proc[j],      true, true);

            if (record)
            {
              try
              {
                fout.write("procedure="+proc[j]);
                fout.newLine();
              }
              catch (Exception e2)
              {
                err=("Error: Problems while recording transformations into file!");
                return -1;
              }
            }

            String x="";
            if (type1[j]==1) x="array=";
            else x="loop=";

            cl1.Add(x+name[j],                 true, true);

            if (record)
            {
              try
              {
                fout.write(x+name[j]);
                fout.newLine();
              }
              catch (Exception e2)
              {
                err=("Error: Problems while recording transformations into file!");
                return -1;
              }
            }
              
            if (!prm2[j].equals(""))
            {
              cl1.Add(prm2[j],                 true, true);

              if (record)
              {
                try
                {
                  fout.write(prm2[j]);
                  fout.newLine();
                }
                catch (Exception e2)
                {
                  err=("Error: Problems while recording transformations into file!");
                  return -1;
                }
              }
            }

            if (corr[j])
            {
              cl1.Add("correct_loops",         true, true);

              if (record)
              {
                try
                {
                  fout.write("correct_loops");
                  fout.newLine();
                }
                catch (Exception e2)
                {
                  err=("Error: Problems while recording transformations into file!");
                  return -1;
                }
              }
            }

            if (type2[j]==0)      x="array_pad="+prm1[j];
            else if (type2[j]==1) x="array_gir="+prm1[j];
            else if (type2[j]==2) x="loop_tiling="+prm1[j];
            else if (type2[j]==3) x="loop_distribution";
            else                  x="loop_unrolling="+prm1[j];
  
            cl1.Add(x,                         true, true);

            if (record)
            {
              try
              {
                fout.write(x);
                fout.newLine();
              }
              catch (Exception e2)
              {
                err=("Error: Problems while recording transformations into file!");
                return -1;
              }
            }
          }
        }

        cl1.Add("go",                          true, true);

        if (record)
        {
          try
          {
            fout.write("go");
            fout.newLine();
          }
          catch (Exception e2)
          {
            err=("Error: Problems while recording transformations into file!");
            return -1;
          }
        }
      }
      else
      {
        cl1.Add("parse",                       true, true);

        if (record)
        {
          try
          {
            fout.write("parse");
            fout.newLine();
          }
          catch (Exception e2)
          {
            err=("Error: Problems while recording transformations into file!");
            return -1;
          }
        }
      }
  
      cl1.Add("terminate",                    false, true);

      if (record)
      {
        try
        {
          fout.write("terminate");
          fout.newLine();
        }
        catch (Exception e2)
        {
          err=("Error: Problems while recording transformations into file!");
          return -1;
        }

        try
        {
          fout.close();
        }
        catch (Exception e5)
        {
          err="Error: Can't close file with recorded transformations!";
          return -1;
        }
      }

      ret=cl1.Send();
      if (ret==-1)
      {
        err="Error: Can't find host with Transformation Server!";
        return -1;
      }
      else if (ret==-2)
      {
        err="Error: Can't connect to Transformation Server!";
        return -1;
      }
      else if (ret==-4)
      {
        err="Error: Can't close connection with Transformation Server!";
        return -1;
      }
      else if (ret==-3)
      {
        err=cl1.GetLastError();
        return -1;
      }
      else if (ret==-5) return 3; //if tr. server returned error (recoverable error)
    }

    //Compile new object files with linking
    //but without error if file was not created
    cl.Add("initiate compilation",   true, true);
    cl.Add("directory="+str_dir,     true, true);
    cl.Add("file_project="+f_prj2,   true, true);
    cl.Add("file_diag_out="+f_tmp,   true, true);
    cl.Add("compilation_mode=1",     true, true);
    cl.Add("no_checking=1",          true, true);
    cl.Add("optimizations="+s_opt,   true, true);
    cl.Add("compile",                true, true);
    cl.Add("terminate",             false, true);

    ret=cl.Send();
    if (ret==-1)
    {
      err="Error: Can't find host with Compilation Server!";
      return -1;
    }
    else if (ret==-2)
    {
      err="Error: Can't connect to Compilation Server!";
      return -1;
    }
    else if (ret==-4)
    {
      err="Error: Can't close connection with Compilation Server!";
      return -1;
    }
    else if (ret<0)
    {
      err=cl.GetLastError();
      return -1;
    }

    //Check here if file was created or not 
    boolean created=true;
    boolean correct=false;

    String x1=str_dir+cp_new.GetExe();

    FileInputStream fin1=null;

    try
    {
      fin1=new FileInputStream(x1);
    }
    catch (FileNotFoundException e1)
    {
      created=false;
    }

    try
    {
      fin1.close();
    }
    catch (Exception e4)
    {}

    if (created)
    {
      //Run transformed application
      cl.Add("initiate compilation",   true, true);
      cl.Add("directory="+str_dir,     true, true);
      cl.Add("file_project="+f_prj2,   true, true);
      cl.Add("file_out_time="+f_tmp1,  true, true);
      cl.Add("file_out_comp="+f_tmp2,  true, true);
      cl.Add("runmode=1",              true, true);
      cl.Add("run",                    true, true);
      cl.Add("terminate",             false, true);

      ret=cl.Send();
      if (ret==-1)
      {
        err="Error: Can't find host with Compilation Server!";
        return -1;
      }
      else if (ret==-2)
      {
        err="Error: Can't connect to Compilation Server!";
        return -1;
      }
      else if (ret==-4)
      {
        err="Error: Can't close connection with Compilation Server!";
        return -1;
      }
      else if (ret<0)
      {
        err=cl.GetLastError();
        return -1;
      }
 
      if (CMisc.isCorrect(str_dir+f_tmp2)==1) 
      {
        correct=true;

        //Get Time
        if (ct.Init(str_dir+f_tmp1)<0)
        {
          err=ct.GetLastError();
          return -1;
        }
      }
      else correct=false;
    }

    //Compile original object files
    cl.Add("initiate compilation",   true, true);
    cl.Add("directory="+str_dir,     true, true);
    cl.Add("file_project="+f_prj1,   true, true);
    cl.Add("file_diag_out="+f_tmp,   true, true);
    cl.Add("compilation_mode=2",     true, true);
    cl.Add("no_checking=1",          true, true);
    cl.Add("optimizations="+s_opt,   true, true);
    cl.Add("compile",                true, true);
    cl.Add("terminate",             false, true);

    ret=cl.Send();
    if (ret==-1)
    {
      err="Error: Can't find host with Compilation Server!";
      return -1;
    }
    else if (ret==-2)
    {
      err="Error: Can't connect to Compilation Server!";
      return -1;
    }
    else if (ret==-4)
    {
      err="Error: Can't close connection with Compilation Server!";
      return -1;
    }
    else if (ret<0)
    {
      err=cl.GetLastError();
      return -1;
    }

    //Return value
    if (!created) return 1; //exe file was not created
                            //probably transformations were not correct


    if (!correct) return 2; //outout files are incorrect
                            //probably transformations were not correct

    return 0;
  }

  //****************************************************************************
  public CTime GetTime()
  {
    CTime ct1=new CTime();
    ct1.Copy(ct);
    return ct1;
  }
}
