//
// eos_cled 
// convert loops with labels into do/enddo format
// for fortran 77 programs
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

//Import libraries...
import java.io.*;

//Main class
public class eos_cled
{
  //**************************************************************************
  public static void main(String args[])
  {
    String str="";
    int i=0;

    //Logo
    System.out.println("EOS Convert loops with labels (f77)  (C) by Grigori G.Fursin   June,2004");

    if (args.length!=2)
    {
      System.out.println("                                             ");
      System.out.println("Usage:                                       ");
      System.out.println(" java cled <input filename> <output filename>");
      System.exit(0);
    }

    //Filename
    int imax=16384;

    int iq=0;

    String[] LabelName=new String [imax];
    int[]    LabelNest=new int [imax];

    String fn1=args[0];
    String fn2=args[1];
  
    BufferedReader fin=null;
    BufferedWriter fout=null;

    try
    {
      fin=new BufferedReader(new FileReader(fn1));
      fout=new BufferedWriter(new FileWriter(fn2));
    }
    catch (IOException e2)
    {
      System.out.println("");
      System.out.println("Fatal Error: Can't open file(s)!");
      System.exit(1);
    }

    try
    {
      while ((str=fin.readLine())!=null)
      {
        String str1=str.trim();
        if (str1.length()>3 && str1.toLowerCase().substring(0,3).equals("do "))
        {
          int il=str1.length();

          String str2=str1.substring(3,il).trim();

          int il1=str2.length();

          if (Character.isDigit(str2.charAt(0)))
          {
            for (i=0; i<il1; i++)
            {
              if (!Character.isDigit(str2.charAt(i))) break;
            }

            int i5=i;

            String str3=str2.substring(0,i5);

            boolean found=false;
            for (i=0; i<iq; i++)
            {
              if (str3.equals(LabelName[i]))
              {
                found=true;
                break;
              }
            }

            if (found)
            {
              LabelNest[i]++;
            }
            else
            {
              LabelName[iq]=str3;
              LabelNest[iq]=0;
              iq++;
            }

            str="      DO "+str2.substring(i5, il1).trim();
          }
        }
        else if (str1.length()>1 && Character.isDigit(str1.charAt(0)))
        {
          int il=str1.length();

          for (i=0; i<il; i++)
          {
            if (!Character.isDigit(str1.charAt(i))) break;
          }

          String str2=str1.substring(0,i);
          String str3=str1.substring(i,il).trim();

          boolean found=false;

          for (i=0; i<iq; i++)
          {
            if (str2.equals(LabelName[i]))
            {
              found=true;
              break;
            }
          }

          if (found)
          {
            if (str3.toLowerCase().equals("continue"))
            {
              str="      END DO";
              for (int i1=0; i1<LabelNest[i]; i1++)
              {
                fout.write(str);
                fout.newLine();
              }
              LabelName[i]="";
              LabelNest[i]=0;
            }
            else
            {
              str="      "+str3;
              fout.write(str);
              fout.newLine();

              boolean finish=false;
              while ((str=fin.readLine())!=null && !finish)
              {

                if (str.length()>5 && str.substring(5,6).equals(">"))
                {
                  fout.write(str);
                  fout.newLine();
                }
                else finish=true;
              }

              for (int i1=0; i1<=LabelNest[i]; i1++)
              {
                fout.write("      END DO");
                fout.newLine();
              }
              LabelName[i]="";
              LabelNest[i]=0;
            }
          }
        }

        fout.write(str);
        fout.newLine();
      }
    }
    catch (Exception e3)
    {
      try
      {
        fin.close();
        fout.close();
      }
      catch (Exception e4)
      {}
      System.out.println("");
      System.out.println("Fatal Error: Problems with file(s)!");
      System.exit(1);
    }

    try
    {
      fin.close();
      fout.close();
    }
    catch (Exception e5)
    {
      System.out.println("");
      System.out.println("Fatal Error: Can't close file(s)!");
      System.exit(1);
    }

    System.out.println("");
    System.out.println("Program finished sucessfully!");

    System.exit(0);
  }
}
